#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
class Routing File
@author: anthony
"""

from netCDF4 import Dataset as NetCDFFile
import numpy as np
import numpy.ma as ma

import matplotlib.pyplot as plt        
import cartopy
import cartopy.crs as ccrs  
import cartopy.io.shapereader as shpreader
from shapely.geometry import Polygon



class routing:
    def __init__(self, dfile):
        self.nc = NetCDFFile(dfile, "r")
        self.varlist = list(self.nc.variables.keys())
         
        self.get_vars()
        self.get_extent()
        self.get_grid()
        self.get_basin_list()    
        
    ##############################
    # Initialization Methods
    #
    def get_vars(self):
        self.var = {}
        for v in self.varlist:
            self.var[v] = self.nc.variables[v][:]
        self.var["basinid"] = ma.masked_where(np.isnan(self.var["basinid"]), self.var["basinid"])
        self.nbpt = int(np.sum(self.var["land"]))
        self.nbasmax = self.var["basinid"].shape[0]
        self.nbnd = self.var["lon_bnd"].shape[0]
        
        self.conv_land2ij = [[np.where(self.var["nbpt_glo"] == i)[0][0], \
                              np.where(self.var["nbpt_glo"] == i)[1][0]] \
                              for i in range(1, self.nbpt+1)]
        self.var_land = {}
        for v in self.varlist:
            if self.var[v].ndim == 2:
                self.var_land[v] = [self.var[v][j,i] for j,i in self.conv_land2ij]
            elif v == "lon_bnd" or v == "lat_bnd":
                self.var_land[v] = [self.var[v][:,j,i] \
                              for j,i in self.conv_land2ij]
            elif self.var[v].ndim == 3:
                self.var_land[v] = [self.var[v][:int(self.var["routenbintobas"][j,i]),j,i] \
                              for j,i in self.conv_land2ij]
        
    def get_grid(self):
        self.GRID = [self.get_polygon(i)  for i in range(self.nbpt)]
    def get_polygon(self, i):
        pts = [(self.var_land["lon_bnd"][i][l], self.var_land["lat_bnd"][i][l]) for l in range(self.nbnd)]
        return Polygon(pts)
    
    def get_extent(self, bid = None):
        self.resh = np.nanmax(np.diff(self.var["lon"][:,:], axis = 1))
        self.resv = np.nanmax(np.diff(self.var["lat"][:,:], axis = 0))
        
        if bid is not None:
            mask = self.basin_mask(bid)
            mask = np.sum(mask, axis = 0)
            A = np.where(mask>0)
            lonnd = [self.var["lon_bnd"][:,A[0][i], A[1][i]] for i in range(len(A[0]))]
            latnd = [self.var["lat_bnd"][:, A[0][i], A[1][i]] for i in range(len(A[0]))]

            ext = [np.min(lonnd)-self.resh, np.max(lonnd)+self.resh, \
                   np.min(latnd)-self.resv, np.max(latnd)+self.resh]
            return ext
        else:
            self.ext = [np.nanmin(self.var["lon_bnd"])-self.resh, np.nanmax(self.var["lon_bnd"])+self.resh, \
                   np.nanmin(self.var["lat_bnd"])-self.resv, np.nanmax(self.var["lat_bnd"])+self.resv]
    ##############################
    # Basic basin operation
    #
    def get_basin_list(self):
        basin_list = ma.masked_where(np.isnan(self.var["basinid"]),self.var["basinid"])
        basin_list = list(ma.unique(basin_list[basin_list.mask == False]))
        area = [np.sum(self.basin_area(bid)) for bid in basin_list]
        self.basin_list = [[int(basin_list[i]), int(area[i])] for i in np.flip(np.argsort(area))] 
        
    def basin_area(self,bid):
        mask = self.basin_mask(bid)
        area = np.nansum(mask*self.var["basin_area"], axis = 0)
        return area
    
    def basin_mask(self,bid):
        A = ma.where(self.var["basinid"] == float(bid))
        pts = self.where2list(A)
        #
        mask = np.zeros(self.var["basinid"].shape)
        mask[A] = 1
        return mask
    
    def basin_listpoints(self, bid):
        mask = np.sum(self.basin_mask(bid), axis = 0)
        lstpts = [i for i in range(self.nbpt) if mask[r1.conv_land2ij[i][0],r1.conv_land2ij[i][1]]> 0]
        return lstpts
    
    def main_fetch(self, bid, p = 90):
        mask = self.basin_mask(bid)
        pts = self.where2list(np.where(mask==1))
        fetch = [self.var["fetch"][a,b,c] for a,b,c in pts]
        #
        maxfetch = np.max(np.array(fetch)) 
        perc = np.percentile(fetch, p)
        A = np.where(fetch>perc)[0]
        pts_out = [pts[i] for i in A]
        return pts_out

    def outflow(self, bid):
        mask = self.basin_mask(bid)
        pts = self.where2list(np.where(mask==1))
        out1 = ma.filled(self.var["routetobasin"], -99)
        outbas1 = np.array([out1[a,b,c] for a,b,c in pts])
        pts_out = [pts[i]for i in np.where(outbas1>self.nbasmax)[0]]
        outflow = [[int(self.var["routetobasin"][a,b,c]), int(self.var["fetch"][a,b,c]),[a,b,c]] for a,b,c in pts_out]

        return outflow

    def inflow_outflow_htus(self, htus):
        x = []
        y = []
        for a,b,c in htus:
            k = int(self.var["routetobasin"][a,b,c]-1)
            if k<self.nbasmax:
               jj,ii = self.conv_land2ij[int(self.var["routetogrid"][a,b,c]-1)]
               x.append((self.var["CG_lon"][a,b,c], self.var["CG_lon"][k,jj,ii]))
               y.append((self.var["CG_lat"][a,b,c], self.var["CG_lat"][k,jj,ii]))
            else:
               x.append([self.var["CG_lon"][a,b,c]])
               y.append([self.var["CG_lat"][a,b,c]])

        return x,y
  
    ##############################
    # Basic Methods
    #  
    def where2list(self, A):
        pts = [[A[0][i],A[1][i], A[2][i]] for i in range(len(A[0]))]
        return pts

