MODULE grid
  !
  USE defprec
  USE haversine
  USE constantes_var
  !
  IMPLICIT NONE
  !
  INTEGER(i_std), SAVE                                     :: NbSegments  !! Number of segments in 
  !                                                                  the polygone defining the grid box
  INTEGER(i_std), SAVE                                     :: NbNeighb    !! Number of neighbours
  !
  REAL(r_std), ALLOCATABLE, DIMENSION (:), SAVE       :: contfrac
  INTEGER(i_std), ALLOCATABLE, DIMENSION (:,:), SAVE  :: neighbours
  REAL(r_std), ALLOCATABLE, DIMENSION(:,:), SAVE      :: headings
  REAL(r_std), ALLOCATABLE, DIMENSION(:,:), SAVE      :: seglength
  REAL(r_std), ALLOCATABLE, DIMENSION(:), SAVE        :: area
  !
  REAL(r_std), ALLOCATABLE, DIMENSION (:,:), SAVE     :: lalo
  !
  INTEGER(i_std), ALLOCATABLE, DIMENSION (:,:), SAVE  :: neighbours_g
  REAL(r_std), ALLOCATABLE, DIMENSION (:,:), SAVE     :: lalo_g
  REAL(r_std), ALLOCATABLE, DIMENSION(:), SAVE        :: area_g
  REAL(r_std), ALLOCATABLE, DIMENSION(:,:,:), SAVE    :: corners_g
  REAL(r_std), ALLOCATABLE, DIMENSION(:,:), SAVE      :: headings_g
  
CONTAINS

  SUBROUTINE grid_init(npts, nbseg, polygons_in, centre_in, area_in, contfrac_in, neighbours_in)

    !! INPUT VARIABLES
    INTEGER(i_std), INTENT(in) :: npts   !! Atmospheric Domain size (unitless)
    INTEGER(i_std), INTENT(in) :: nbseg  !! Number of segments in the polygone
    REAL(r_std), INTENT(in)    :: polygons_in(npts,2*nbseg+1,2)
    REAL(r_std), INTENT(in)    :: centre_in(npts,2)
    REAL(r_std), INTENT(in)    :: contfrac_in(npts)
    REAL(r_std), INTENT(in)    :: area_in(npts)
    INTEGER, INTENT(in) :: neighbours_in(npts,nbseg*2)
    !
    !
    !! Local variables
    INTEGER(i_std) :: i, is, iss
    REAL(r_std), DIMENSION(npts,2*nbseg) :: lonpoly, latpoly
    LOGICAL :: debug = .FALSE.
    !
    ! Define dimension of polygones
    !
    NbSegments = nbseg
    NbNeighb = 2*NbSegments
    !
    ! Allocate memory and attribute values
    !
    IF ( .NOT. ALLOCATED(contfrac) ) THEN
       ALLOCATE(contfrac(npts))
       contfrac(:) = contfrac_in(:)
    ENDIF
    IF ( (.NOT.ALLOCATED(neighbours))) THEN
       ALLOCATE(neighbours(npts,NbNeighb))
       neighbours(:,:) = neighbours_in
    ENDIF
    IF ( (.NOT.ALLOCATED(neighbours_g))) THEN
       ALLOCATE(neighbours_g(npts,NbNeighb))
       neighbours_g(:,:) = neighbours_in
    ENDIF
    IF ( (.NOT.ALLOCATED(headings))) THEN
       ALLOCATE(headings(npts,NbNeighb))
       headings(:,:) = val_exp
    ENDIF
    IF ( (.NOT.ALLOCATED(headings_g))) THEN
       ALLOCATE(headings_g(npts,NbNeighb))
       headings_g(:,:) = val_exp
    ENDIF
    IF ( (.NOT.ALLOCATED(seglength))) THEN
       ALLOCATE(seglength(npts,NbSegments))
       seglength(:,:) = val_exp
    ENDIF
    IF ( (.NOT.ALLOCATED(corners_g))) THEN
       ALLOCATE(corners_g(npts,NbSegments,2))
       corners_g(:,:,:) = val_exp
    ENDIF
    IF ( (.NOT.ALLOCATED(area))) THEN
       ALLOCATE(area(npts))
       area(:) = area_in(:)
    ENDIF
    IF ( (.NOT.ALLOCATED(area_g))) THEN
       ALLOCATE(area_g(npts))
       area_g(:) = area_in(:)
    ENDIF
    IF ( (.NOT.ALLOCATED(lalo))) THEN
       ALLOCATE(lalo(npts,2))
       lalo(:,:) = val_exp
    ENDIF
    IF ( (.NOT.ALLOCATED(lalo_g))) THEN
       ALLOCATE(lalo_g(npts,2))
       lalo_g(:,:) = val_exp
    ENDIF
    !
    ! Compute other variables we might need
    !
    !
    ! Save the longitude and latitudes of the nbseg corners_g (=vertices)
    !
    DO i=1,npts
       lonpoly(i,:) = polygons_in(i,1:2*nbseg,1)
       latpoly(i,:) = polygons_in(i,1:2*nbseg,2)
       !
       lalo(i,1) = centre_in(i,2)
       lalo(i,2) = centre_in(i,1)
       lalo_g(i,1) = centre_in(i,2)
       lalo_g(i,2) = centre_in(i,1)
    ENDDO
    !
    ! Get the heading normal to the 4 segments and through the 4 corners_g.
    !
    CALL haversine_polyheadings(npts, nbseg, lonpoly, latpoly, centre_in, headings)
    !
    ! Order the points of the polygone in clockwise order Starting with the northern most
    !
    CALL haversine_polysort(npts, nbseg, lonpoly, latpoly, headings, neighbours)
    !
    DO i=1,npts
       DO is=1,nbseg
          iss=(is-1)*2+1
          corners_g(i,is,1) = lonpoly(i,iss)
          corners_g(i,is,2) = latpoly(i,iss)
       ENDDO
       headings_g(i,:) =  headings(i,:)
    ENDDO
    !
    IF ( debug ) THEN
       DO is=1,npts
          CALL grid_diag(is, npts, centre_in)
       ENDDO
    ENDIF
    !
  END SUBROUTINE grid_init

  SUBROUTINE grid_diag(is, npts, center)
    !
    !
    INTEGER(i_std), INTENT(in) :: is, npts
    REAL(r_std), INTENT(in)    :: center(npts,2)
    !
    CHARACTER(LEN=25) :: NW, NN, NE, EE, SE, SS, SW, WW, blk
    !
    blk = "              --       "
    IF (neighbours(is,8) > 0) THEN
       WRITE(NW,'(2F12.3)') center(neighbours(is,8),:)
    ELSE
       NW=blk
    ENDIF
    IF (neighbours(is,1) > 0) THEN
       WRITE(NN,'(2F12.3)') center(neighbours(is,1),:)
    ELSE
       NN=blk
    ENDIF
    IF (neighbours(is,2) > 0) THEN
       WRITE(NE,'(2F12.3)') center(neighbours(is,2),:)
    ELSE
       NE=blk
    ENDIF
    IF (neighbours(is,3) > 0) THEN
       WRITE(EE,'(2F12.3)') center(neighbours(is,3),:)
    ELSE
       EE=blk
    ENDIF    
    IF (neighbours(is,4) > 0) THEN
       WRITE(SE,'(2F12.3)') center(neighbours(is,4),:)
    ELSE
       SE=blk
    ENDIF
    IF (neighbours(is,5) > 0) THEN
       WRITE(SS,'(2F12.3)') center(neighbours(is,5),:)
    ELSE
       SS=blk
    ENDIF
    IF (neighbours(is,6) > 0) THEN
       WRITE(SW,'(2F12.3)') center(neighbours(is,6),:)
    ELSE
       SW=blk
    ENDIF
    IF (neighbours(is,7) > 0) THEN
       WRITE(WW,'(2F12.3)') center(neighbours(is,7),:)
    ELSE
       WW=blk
    ENDIF
    WRITE(*,*) "AROUND : ", center(is,:)
    WRITE(*,*) "---------------------------------------------------------------"
    WRITE(*,*) NW," <--> ",       NN, " <--> ",            NE
    WRITE(*,*) WW,"         XXXXXXXXXXXXX               ", EE
    WRITE(*,*) SW," <--> ",       SS, " <--> ",            SE
    WRITE(*,*) "---------------------------------------------------------------"
    WRITE(*,*) headings(is,8)," <--> ",headings(is,1)," <--> ",  headings(is,2)
    WRITE(*,*) headings(is,7)," <--> 000000000000000000 <--> ",  headings(is,3)
    WRITE(*,*) headings(is,6)," <--> ",headings(is,5)," <--> ",  headings(is,4)
    WRITE(*,*) "---------------------------------------------------------------"
    
  END SUBROUTINE grid_diag
  
END MODULE grid
