!  ==============================================================================================================================\n
!  MODULE 	: This module provides a few basic operations for convex polygons on the sphere and in longitude
!                 latitude coordinates. These operations are based on the haversine equations for great circle arcs.
!
!                 Definition of polygons : the basic assumption is that they describe grid boxes and are thus 
!                 provided with more points than strictly necessary. Each polygon starts at index 1 on a vertex and
!                 alternates mid-points of segments and vertices. The mid-point of segments are kept as they are
!                 useful elements for the operations on the grid boxes.
!
!                 The module provides the following subroutine and functions :
!                 REMOVED : haversine_reglatlontoploy : Lists the polygons and all their attributes (centre point, 
!                                             neighbouring grids, indices in i and j on the original grid) for 
!                                             regular lon/lat grid.
!                 REMOVED : haversine_regxytoploy : As above but for grid boxes on the sphere projected onto a regular
!                                         X/Y grid.
!                 REMOVED : haversine_singlepointpoly : Computes the polygone around a given point with a known area.
!                 haversine_polyheadings : Compute the heading out of the polygon for each vertex and mid-point of
!                                          segment.
!                 haversine_polysort : Sort the polygons so that all points are in clockwise order.
!                 haversine_polyseglen : Compute the length of all segments of the polygon using great circles.
!                 haversine_polyseglen : Compute the length of all segments of the polygon on a regular lat lon grid.
!                 haversine_clockwise : Get the indices which sort a polygon in a clockwise order starting from an
!                                       initial vertex.
!                 haversine_polyarea : Computes the area covered by a polygon on the sphere.
!                 haversine_laloarea : Compute the area for a regular lat lon grid.
!                 haversine_xyarea : Compute the area for the special case where the grid box size in x and y are already 
!                                    given by the projection.
!                 haversine_heading : Initial heading between start point and end point along a great circle.
!                 haversine_distance : Compute the distance between 2 points along the great circle.
!                 haversine_diffangle : Computes the angle between two headings.
!                 haversine_radialdis : Compute the coordinates found in a given heading and distance.
!                 haversine_dtor : Degrees to radians
!                 haversine_rtod : Radians to degrees
!
!  CONTACT      : jan.polcher@lmd.jussieu.fr
!
!  LICENCE      : IPSL (2016)
!  This software is governed by the CeCILL licence see ORCHIDEE/ORCHIDEE_CeCILL.LIC
!
!>\BRIEF        
!! 
!! RECENT CHANGE(S): None 
!! 
!! REFERENCE(S) : None
!! 
!_ ================================================================================================================================
MODULE haversine

  USE defprec
  USE ioipsl
  USE constantes_var
!!  USE module_llxy

  IMPLICIT NONE
  !
  !
  PRIVATE
  PUBLIC :: haversine_polyheadings, haversine_polysort, &
       &    haversine_polyseglen, haversine_laloseglen, haversine_clockwise, haversine_polyarea, &
       &    haversine_laloarea, haversine_xyarea, haversine_diffangle, haversine_heading

CONTAINS
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_polyheadings   
!!
!>\BRIEF       Compute the heading out of the polygon for each vertex and mid-point of
!              segment.
!!
!! DESCRIPTION: This heading is computed by using the great circle between the centre of the polygon and
!!              the point on the boundary considered. The direction is the one facing outwards from the polygon.
!!
!! \n
!_ ==============================================================================================================================
!!
!!
  SUBROUTINE haversine_polyheadings(nbpt, nbseg, lonpoly, latpoly, center, headings)
    !
    ! 0.1 Input variables 
    ! Size of the gathered domain
    INTEGER(i_std), INTENT(in)                                 :: nbpt
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    !
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: lonpoly
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: latpoly
    REAL(r_std), DIMENSION(nbpt,2), INTENT(in)                 :: center
    !
    ! 0.2 Output variables
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(out)          :: headings
    !
    ! 0.3 Local variables
    !
    INTEGER(i_std) :: i, ns
    !
    ! We compute for each vertex of our polygon (actual vertices and mid-points) the direction to the
    ! centre. The we add 180. to get the opposite direction.
    !
    DO i=1,nbpt
       DO ns=1,nbseg*2
          headings(i,ns) = MOD(haversine_heading(lonpoly(i,ns), latpoly(i,ns), center(i,1), center(i,2))+180.0, 360.0)
       ENDDO
    ENDDO
    !
  END SUBROUTINE haversine_polyheadings
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_polysort     
!!
!>\BRIEF Sort the polygons so that all points are in clockwise order.	  
!!
!! DESCRIPTION:	  
!!
!! \n
!_ ==============================================================================================================================
!!
  SUBROUTINE haversine_polysort(nbpt, nbseg, lonpoly, latpoly, headings, neighb)
    !
    ! This subroutine is foreseen for polygones which start with a vertex and then alternate
    ! with mid-points of the segments. The heading at a vertex is the direction (along the 
    ! great circle) between the center of the polygone and this vertex. At a segment mid-point
    ! the direction is the normal facing outward.
    !
    ! 0.1 Input Variables
    ! Size of the gathered domain
    INTEGER(i_std), INTENT(in)                                 :: nbpt
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    !
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(inout)        :: lonpoly
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(inout)        :: latpoly
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(inout)        :: headings
    INTEGER(i_std), DIMENSION(nbpt,nbseg*2), INTENT(inout)     :: neighb
    !
    ! 0.2 Local variables
    !
    INTEGER(i_std) :: i, ic, startindex
    INTEGER(i_std), DIMENSION(nbseg*2)     :: reindex
    REAL(r_std)                            :: starthead
    REAL(r_std), DIMENSION(nbseg*2)        :: lon_loc
    REAL(r_std), DIMENSION(nbseg*2)        :: lat_loc
    REAL(r_std), DIMENSION(nbseg*2)        :: head_loc
    REAL(r_std), DIMENSION(nbseg*2)        :: negb_loc
    !
    DO i=1,nbpt
       head_loc(:) = headings(i,:)
       lon_loc(:) = lonpoly(i,:)
       lat_loc(:) = latpoly(i,:)
       negb_loc(:) = neighb(i,:)
       !
       ! The first vertice of our polygone needs to be the heading closest to 
       ! North (0 degree). No difference is done between vertices and mid-points
       ! of segments.
       !
       starthead=0.0
       !
       CALL haversine_clockwise(nbseg, head_loc, starthead, reindex)
       !
       !
       headings(i,:) = head_loc(reindex(:))
       lonpoly(i,:) = lon_loc(reindex(:))
       latpoly(i,:) = lat_loc(reindex(:))
       neighb(i,:) = negb_loc(reindex(:))
       !
    ENDDO
    !
  END SUBROUTINE haversine_polysort
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_polyseglen   
!!
!>\BRIEF       Compute the length of all segments of the polygon using the great circle.
!!
!! DESCRIPTION:	  
!!
!! \n
!_ ==============================================================================================================================
!!
  SUBROUTINE haversine_polyseglen(nbpt, nbseg, lonpoly, latpoly, seglength)
    !
    ! Computes the segment length for each of the polygones. These are
    ! polygones with middle points given for each segment. This we need
    ! to take only every other point.
    !
    !
    ! 0.1 Input Variables
    ! Size of the gathered domain
    INTEGER(i_std), INTENT(in)                                 :: nbpt
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    !
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: lonpoly
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: latpoly
    REAL(r_std), DIMENSION(nbpt,nbseg), INTENT(out)            :: seglength
    !
    ! 0.2 Local variables
    !
    INTEGER(i_std) :: i, iv, istart, iend, iseg, ioff
    REAL(r_std)    :: slpm1, slpp1
    !
    DO i=1,nbpt
       iseg = 0
       !
       ! Find if the first element is a vertex or mid-point of segment.
       ! Get the 2 headings out of the start point. If these headings are larger than 135 degrees
       ! then probably this point is a segment mid-point.
       !
       slpm1 = haversine_heading(lonpoly(i,1), latpoly(i,1), lonpoly(i,nbseg*2), latpoly(i,nbseg*2))
       slpp1 = haversine_heading(lonpoly(i,1), latpoly(i,1), lonpoly(i,2), latpoly(i,2))
       !
       IF ( ABS(MOD(slpp1-slpm1, 360.0)) > 135.0 ) THEN
          ! The polygon starts with a segment mid-point
          ioff = -1
       ELSE
          ioff = 0
       ENDIF
       !
       DO iv=1,nbseg*2,2
          !
          istart=MODULO((iv-1)+ioff,nbseg*2)+1
          iend=MODULO((iv-1)+ioff+2,nbseg*2)+1
          iseg = iseg + 1
          !
          seglength(i,iseg) = haversine_distance(lonpoly(i,istart), latpoly(i,istart), &
               &                                 lonpoly(i,iend), latpoly(i,iend))
       ENDDO
    ENDDO
    !
  END SUBROUTINE haversine_polyseglen
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_laloseglen   
!!
!>\BRIEF       Compute the length of all segments of the polygon when on a regular Lat Lon grid.
!!
!! DESCRIPTION:	  
!!
!! \n
!_ ==============================================================================================================================
!!
  SUBROUTINE haversine_laloseglen(nbpt, nbseg, lonpoly, latpoly, seglength)
    !
    ! Computes the segment length for each of the polygones. These are
    ! polygones with middle points given for each segment. This we need
    ! to take only every other point.
    !
    !
    ! 0.1 Input Variables
    ! Size of the gathered domain
    INTEGER(i_std), INTENT(in)                                 :: nbpt
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    !
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: lonpoly
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: latpoly
    REAL(r_std), DIMENSION(nbpt,nbseg), INTENT(out)            :: seglength
    !
    ! 0.2 Local variables
    !
    INTEGER(i_std) :: i, iv, istart, iend, ioff, iseg
    REAL(r_std)    :: coslat, slpm1, slpp1
    !
    DO i=1,nbpt
       iseg = 0
       !
       ! Find if the first element is a vertex or mid-point of segment.
       ! Get the 2 headings out of the start point. If these headings are larger than 135 degrees
       ! then probably this point is a segment mid-point.
       !
       slpm1 = haversine_heading(lonpoly(i,1), latpoly(i,1), lonpoly(i,nbseg*2), latpoly(i,nbseg*2))
       slpp1 = haversine_heading(lonpoly(i,1), latpoly(i,1), lonpoly(i,2), latpoly(i,2))
       !
       IF ( ABS(MOD(slpp1-slpm1, 360.0)) > 135.0 ) THEN
          ! The polygon starts with a segment mid-point
          ioff = -1
       ELSE
          ioff = 0
       ENDIF
       !
       DO iv=1,nbseg*2,2
          !
          istart=MODULO((iv-1)+ioff,nbseg*2)+1
          iend=MODULO((iv-1)+ioff+2,nbseg*2)+1
          iseg = iseg + 1
          !
          !
          IF ( ABS(lonpoly(i,istart)-lonpoly(i,iend)) < EPSILON(lonpoly) ) THEN
             !
             ! Distance along a meridian
             !
             seglength(i,iseg) =  ABS(latpoly(i,istart) - latpoly(i,iend)) * &
                     pi/180. * R_Earth
             !
          ELSE IF ( ABS(latpoly(i,istart)-latpoly(i,iend)) < EPSILON(latpoly) ) THEN
             !
             ! Distance along a circle of constant latitude
             !
             coslat = MAX(COS(latpoly(i,istart) * pi/180.), mincos)
             seglength(i,iseg) = ABS( lonpoly(i,istart) - lonpoly(i,iend) ) * &
                  pi/180. * R_Earth * coslat
             !
          ELSE
             CALL ipslerr(3, "haversine_laloseglen", "The polygon here does not originate from a regular", &
               &       "latitude longitude grid.","")
          ENDIF
          !
       ENDDO
    ENDDO
    !
  END SUBROUTINE haversine_laloseglen
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_clockwise   
!!
!>\BRIEF       Get the indices which sort a polygon in a clockwise order starting from an
!              initial vertex given by start.	  
!!
!! DESCRIPTION:	  
!!
!! \n
!_ ==============================================================================================================================
!!
  SUBROUTINE haversine_clockwise(nbseg, heading, start, sortindex)
    !
    ! Find the order of the polygone vertices which start at "start" and 
    ! follow in a clockwise direction.
    !
    ! 0.1 Input Variables
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    REAL(r_std), DIMENSION(nbseg*2), INTENT(in)                :: heading
    REAL(r_std), INTENT(in)                                    :: start
    ! 
    ! 0.2 Output variable
    !
    INTEGER(i_std), DIMENSION(nbseg*2), INTENT(out)            :: sortindex
    !
    ! 0.3 Local variables
    !
    INTEGER(i_std) :: is, js, imin(1)
    REAL(r_std)    :: delang
    REAL(r_std)    :: undef = 9999999999.99999
    REAL(r_std), DIMENSION(nbseg*2) :: workhead
    !
    delang = 360.0/(nbseg*2)
    !
    workhead(:) = heading(:)
    !
    DO is=1,nbseg*2
       !
       ! Compute the difference of heading to the next target angle
       !
       DO js=1,nbseg*2
          IF ( workhead(js) < undef ) THEN
             workhead(js) = MOD(heading(js)-(start+(is-1)*delang)+360.0, 360.0)
             ! Transfer to -180:180 interval
             IF (workhead(js) > 180.0) workhead(js)=workhead(js)-360.0
          ENDIF
       ENDDO
       !
       ! Locate the vertex closest to that target angle
       !
       imin=MINLOC(ABS(workhead))
       sortindex(is) = imin(1)
       !
       ! Mask this vertex so that it is skipped in the next iteration
       !
       workhead(imin(1)) = undef
       !
    ENDDO
    !
  END SUBROUTINE haversine_clockwise
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_polyarea   
!!
!>\BRIEF       Computes the area covered by a polygon on the sphere.	  
!!
!! DESCRIPTION:	Computes the area of each polygon based on Girard's theorem. It
!!              states that the area of a polygon of great circles is R**2 times 
!!              the sum of the angles between the polygons minus (N-2)*pi where N 
!!             x is number of corners.  
!!
!! \n
!_ ==============================================================================================================================
!!
  SUBROUTINE haversine_polyarea(nbpt, nbseg, lonpoly, latpoly, area)
    !
    !
    ! 0.1 Input Variables
    ! Size of the gathered domain
    INTEGER(i_std), INTENT(in)                                 :: nbpt
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    !
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: lonpoly
    REAL(r_std), DIMENSION(nbpt,nbseg*2), INTENT(in)           :: latpoly
    REAL(r_std), DIMENSION(nbpt), INTENT(out)                  :: area
    !
    ! 0.2 Local variables
    !
    INTEGER(i_std) :: i, ia, ib1, ib2, iseg
    REAL(r_std)    :: beta1, beta2
    REAL(r_std), DIMENSION(nbseg) :: angles
    !
    DO i=1,nbpt
       iseg=0
       DO ia=1,nbseg*2,2
          !! Index of next vertex, i.e. ia+1
          ib1=MOD(ia+1, nbseg*2)+1
          !! Index of previous vertex, i.e. ia-2
          ib2=MOD(ia+nbseg*2-3, nbseg*2)+1
          iseg=iseg+1
          !
          beta1=haversine_dtor(haversine_heading(lonpoly(i,ia), latpoly(i,ia), lonpoly(i,ib1), latpoly(i,ib1)))
          beta2=haversine_dtor(haversine_heading(lonpoly(i,ia), latpoly(i,ia), lonpoly(i,ib2), latpoly(i,ib2)))
          !
          angles(iseg)=acos(cos(-beta1)*cos(-beta2) + sin(-beta1)*sin(-beta2))
          !
       ENDDO
       area(i) = (sum(angles) - (nbseg-2)*pi)*R_Earth**2
    ENDDO
    !
  END SUBROUTINE haversine_polyarea
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_laloarea   
!!
!>\BRIEF       Computes the area of a regular latitude longitude box for which we already have the
!!             the segment length.	  
!!
!! DESCRIPTION:	Just verify that we have 4 segments. 
!!
!! \n
!_ ==============================================================================================================================
!!
  SUBROUTINE haversine_laloarea(nbpt, nbseg, seglen, area)
    !
    !
    ! 0.1 Input Variables
    ! Size of the gathered domain
    INTEGER(i_std), INTENT(in)                                 :: nbpt
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    !
    REAL(r_std), DIMENSION(nbpt,nbseg), INTENT(in)             :: seglen
    REAL(r_std), DIMENSION(nbpt), INTENT(out)                  :: area
    !
    ! 0.2 Local variables
    !
    INTEGER(i_std) :: i
    !
    IF ( nbseg .NE. 4 ) THEN
          CALL ipslerr(3, "haversine_laloarea", "We need to have 4 segments in the polygons", &
               &       "to use this subroutine","")
    ENDIF
    !
    DO i=1,nbpt
       area(i) = (seglen(i,1)+seglen(i,3))/2.0*(seglen(i,2)+seglen(i,4))/2.0
    ENDDO
    !
  END SUBROUTINE haversine_laloarea
!!
!!  =============================================================================================================================
!! SUBROUTINE: haversine_laloarea   
!!
!>\BRIEF       Computes the area in the special case where the projection has given us the size in X and Y of each
!!             grid box.
!!
!! DESCRIPTION:
!!
!! \n
!_ ==============================================================================================================================
!!
  SUBROUTINE haversine_xyarea(nbpt, nbseg, ilandtoij, jlandtoij, dx, dy, area)
    !
    !
    ! 0.1 Input Variables
    ! Size of the gathered domain
    INTEGER(i_std), INTENT(in)                                 :: nbpt
    ! Number of segments
    INTEGER(i_std), INTENT(in)                                 :: nbseg
    !
    INTEGER(i_std), DIMENSION(nbpt), INTENT(in)                :: ilandtoij, jlandtoij
    REAL(r_std), DIMENSION(:,:), INTENT(in)                    :: dx, dy
    REAL(r_std), DIMENSION(nbpt), INTENT(out)                  :: area
    !
    ! 0.2 Local variables
    !
    INTEGER(i_std) :: il
    !
    DO il=1,nbpt
       area(il) = dx(ilandtoij(il),jlandtoij(il))*dy(ilandtoij(il),jlandtoij(il))
    ENDDO
    !
  END SUBROUTINE haversine_xyarea
!!
!!  =============================================================================================================================
!! FUNCTIONS: haversine_heading,  haversine_distance, haversine_dtor, haversine_rtod 
!!
!>\BRIEF Various functions to help with the calculations in this module.	  
!!
!! DESCRIPTION:	  
!!
!! \n
!_ ==============================================================================================================================
  REAL(r_std) FUNCTION haversine_heading(lon_start, lat_start, lon_end, lat_end)
    !
    ! The heading is an angle in degrees between 0 and 360.
    !
    IMPLICIT NONE
    REAL(r_std), INTENT(IN) :: lon_start, lat_start, lon_end, lat_end
    !
    REAL(r_std) :: dlon, lat1, lat2, y, x
    !
    dlon = haversine_dtor(lon_end-lon_start)
    lat1 = haversine_dtor(lat_start)
    lat2 = haversine_dtor(lat_end)
    y = sin(dlon) * cos(lat2)
    x = cos(lat1)* sin(lat2) - sin(lat1) * cos(lat2)* cos(dLon)
    haversine_heading = MOD(haversine_rtod(atan2(y,x))+360.0, 360.0)
    !
  END FUNCTION haversine_heading
  !!
  !!
  !!
  REAL(r_std) FUNCTION haversine_diffangle(x,y)
    !
    ! This function compues the smallest angle between 2 headings.
    ! A positive angle will be in mathematical positive direction (anti-clockwise)
    ! and a negative one is measured clockwise.
    !
    IMPLICIT NONE
    REAL(r_std), INTENT(IN) :: x, y
    !
    REAL(r_std) :: rx, ry
    !
    rx=haversine_dtor(x)
    ry=haversine_dtor(y)
    haversine_diffangle = haversine_rtod(atan2(sin(ry-rx), cos(ry-rx)))
    !
  END FUNCTION haversine_diffangle
  !!
  !!
  !!
  REAL(r_std) FUNCTION haversine_distance(lon_start, lat_start, lon_end, lat_end)
    IMPLICIT NONE
    REAL(r_std), INTENT(IN) :: lon_start, lat_start, lon_end, lat_end
    !
    REAL(r_std) :: dlon, dlat, lat1, lat2, a, c
    !
    dlat = haversine_dtor(lat_end-lat_start)
    dlon = haversine_dtor(lon_end-lon_start)
    lat1 = haversine_dtor(lat_start)
    lat2 = haversine_dtor(lat_end)
    a = sin(dlat/2) * sin(dlat/2) + sin(dlon/2) * sin(dlon/2) * cos(lat1) * cos(lat2)
    c = 2 * atan2(sqrt(a), sqrt(1-a))
    haversine_distance = c * R_Earth
    !
  END FUNCTION haversine_distance
  !!
  !!
  !!
  FUNCTION haversine_radialdis(lon_start, lat_start, head, dis)
    IMPLICIT NONE
    REAL(r_std), INTENT(IN)   :: lon_start, lat_start, head, dis
    REAL(r_std), DIMENSION(2) :: haversine_radialdis
    !
    REAL(r_std) :: lonout, latout, lat1, lon1, lat, lon, dlon, tc
    !
    lon1 = haversine_dtor(lon_start)
    lat1 = haversine_dtor(lat_start)
    tc = haversine_dtor(head)
    !
    lat =asin(sin(lat1)*cos(dis/R_Earth)+cos(lat1)*sin(dis/R_Earth)*cos(tc))
    dlon = atan2(sin(tc)*sin(dis/R_Earth)*cos(lat1),cos(dis/R_Earth)-sin(lat1)*sin(lat))
    lon = MOD(lon1+dlon+pi, 2*pi)-pi
    !
    latout = haversine_rtod(lat)
    lonout = haversine_rtod(lon)
    !
    haversine_radialdis(1) = lonout
    haversine_radialdis(2) = latout
    !
  END FUNCTION haversine_radialdis
  !
  ! --------------------------------------------------------------------
  ! FUNCTION  haversine_dtor():
  !    This function takes a REAL argument in degree and converts it to
  ! the equivalent radian.
  ! --------------------------------------------------------------------
  REAL(r_std) FUNCTION  haversine_dtor(Degree)
    IMPLICIT  NONE
    REAL(r_std), INTENT(IN) :: Degree
    haversine_dtor = Degree * pi/180.0
  END FUNCTION haversine_dtor
  ! --------------------------------------------------------------------
  ! FUNCTION  haversine_rtod():
  !    This function takes a REAL argument in radian and converts it to
  ! the equivalent degree.
  ! --------------------------------------------------------------------
  REAL(r_std) FUNCTION  haversine_rtod(Radian)
    IMPLICIT  NONE
    REAL(r_std), INTENT(IN) :: Radian

    haversine_rtod = Radian * 180.0/pi
  END FUNCTION haversine_rtod
  !!
  !!
  !!
END MODULE haversine
