#
# Produce a lambert Conformal grid as available in WRF
#
import numpy as np
from numba import jit
#
import configparser
config=configparser.ConfigParser()
config.read("run.def")
#
EarthRadius=config.getfloat("OverAll", "EarthRadius")
rad_per_deg = np.pi/180.
deg_per_rad = 180./np.pi
#
@jit
def cone(truelat1, truelat2) :
    if (np.abs(truelat1-truelat2) > 0.1) :
        cone = np.log10(np.cos(truelat1*rad_per_deg)) - np.log10(np.cos(truelat2*rad_per_deg))
        cone = cone / (np.log10(np.tan((45.0 - np.abs(truelat1)/2.0) * rad_per_deg)) - np.log10(np.tan((45.0 - np.abs(truelat2)/2.0) * rad_per_deg)))
    else :
        cone = np.sin(np.abs(truelat1)*rad_per_deg )
    return cone
#
# Lambert Conformal projection
#
@jit
def LC_ijll(polyij, hemi, truelat1, truelat2, polei, polej, rebydx, stdlon, cone) :
            
    polyll=[]
    
    chi1 = (90. - hemi*truelat1)*rad_per_deg
    chi2 = (90. - hemi*truelat2)*rad_per_deg
    
    for i,j in polyij :
        
        inew = hemi * i
        jnew = hemi * j
        
        xx = inew - polei
        yy = polej - jnew
        r2 = (xx*xx + yy*yy)
        r = np.sqrt(r2)/rebydx
        
        if (r2 == 0.) :
            lat = hemi * 90.
            lon = stdlon
        else :
            lon = stdlon + deg_per_rad * np.arctan2(hemi*xx,yy)/cone
            lon = np.mod(lon+360., 360.)
            
            if (chi1 == chi2) :
                chi = 2.0*np.arctan( ( r/np.tan(chi1) )**(1./cone) * np.tan(chi1*0.5) )
            else :
                chi = 2.0*np.arctan( (r*cone/np.sin(chi1))**(1./cone) * np.tan(chi1*0.5))
                
            lat = (90.0-chi*deg_per_rad)*hemi
                    
            if (lon > +180.) : lon = lon - 360.
            if (lon < -180.) : lon = lon + 360.

        polyll.append([lon,lat])

    return polyll

@jit
def LC_llij(polyll, hemi, truelat1, truelat2, polei, polej, rebydx, stdlon, cone) :
            
    polyij=[]

    for lon,lat in polyll :
        
        deltalon = lon - stdlon
        if (deltalon > +180.) : deltalon = deltalon - 360.
        if (deltalon < -180.) : deltalon = deltalon + 360.
        
        tl1r = truelat1 * rad_per_deg
        ctl1r = np.cos(tl1r)     
        
        rm = rebydx * ctl1r/cone * (np.tan((90.*hemi-lat)*rad_per_deg/2.) / np.tan((90.*hemi-truelat1)*rad_per_deg/2.))**cone
        
        arg = cone*(deltalon*rad_per_deg)
        i = polei + hemi * rm * np.sin(arg)
        j = polej - rm * np.cos(arg)
        
        i = int(np.rint(hemi * i))
        j = int(np.rint(hemi * j))
        
        polyij.append([i,j])

    return polyij

class LambertC:
    def __init__(self, dx, known_lon, known_lat, truelat1, truelat2, stdlon):
        self.dx = dx
        self.truelat1 = truelat1
        self.truelat2 = truelat2
        if (np.abs(truelat2) > 90. ) : self.truelat2 = truelat1
        self.lon1 = known_lon
        self.lat1 = known_lat
        self.knowni = 1
        self.knownj = 1
        self.stdlon = stdlon
        
        if truelat1 < 0. :
            self.hemi = -1.0
        else:
            self.hemi = 1.0
        self.rebydx = EarthRadius/dx
        self.cone = cone(truelat1, truelat2)

        deltalon1 = self.lon1 - self.stdlon
        if (deltalon1 > +180.) : deltalon1 = deltalon1 - 360.
        if (deltalon1 < -180.) : deltalon1 = deltalon1 + 360.
        tl1r = self.truelat1 * rad_per_deg
        ctl1r = np.cos(tl1r)
        self.rsw = self.rebydx * ctl1r/self.cone * (np.tan((90.*self.hemi-self.lat1)*rad_per_deg/2.) / np.tan((90.*self.hemi-self.truelat1)*rad_per_deg/2.))**self.cone
  
        arg = self.cone*(deltalon1*rad_per_deg)
        self.polei = self.hemi*self.knowni - self.hemi * self.rsw * np.sin(arg)
        self.polej = self.hemi*self.knownj + self.rsw * np.cos(arg)
        
    def ijll(self,polyij) :

        polyll = LC_ijll(polyij, self.hemi, self.truelat1, self.truelat2, self.polei, self.polej, self.rebydx, self.stdlon, self.cone)
                
        return polyll

    def llij(self, polyll) :

        polyij = LC_llij(polyll, self.hemi, self.truelat1, self.truelat2, self.polei, self.polej, self.rebydx, self.stdlon, self.cone)
                
        return polyij
    
##################################################################
#
# Regular Lon/Lat projjection
#
##################################################################

@jit
def RLL_ijll(polyij, dlon, dlat, ilon, ilat) :
    polyll=[]

    for i,j in polyij :

        lon = ilon+(i-1)*dlon
        lat = ilat+(j-1)*dlat
        if (lon > +180.) : lon = lon - 360.
        if (lon < -180.) : lon = lon + 360.
        polyll.append([lon,lat])
            
    return polyll

@jit
def RLL_llij(polyll, dlon, dlat, ilon, ilat) :
    polyij=[]
    
    for lon,lat in polyll :
        
        i = int(np.rint((lon-ilon)/dlon)+1)
        j = int(np.rint((lat-ilat)/dlat)+1)
        polyij.append([i,j])

    return polyij

class RegLonLat :
    def __init__(self, dlon, dlat, ilon, ilat) :
        self.dlon = dlon
        self.dlat = dlat
        self.initlon = ilon
        self.initlat = ilat

    def ijll(self,polyij) :
        
        polyll = RLL_ijll(polyij, self.dlon, self.dlat, self.initlon, self.initlat)        
            
        return polyll
    
    def llij(self, polyll) :

        polyij = RLL_llij(polyll, self.dlon, self.dlat, self.initlon, self.initlat)
            
        return polyij 
