#
import numpy as np
from numba import jit
import os, sys
#
FillValue=np.nan
IntFillValue=-999999
vtyp=np.float64
vinttyp=np.int32
#
from netCDF4 import Dataset
import pickle
from spherical_geometry import vector
import time
#
# Configuration
#
import configparser
config=configparser.ConfigParser()
config.read("run.def")
saveweights=config.get("OverAll", "SaveWeights", fallback="true")
EarthRadius=config.getfloat("OverAll", "EarthRadius", fallback=6370000.0)
#
# Logging in MPI : https://groups.google.com/forum/#!topic/mpi4py/SaNzc8bdj6U
#
import getargs
log_master, log_world = getargs.getLogger()
INFO, DEBUG, ERROR = log_master.info, log_master.debug, log_world.error
INFO_ALL, DEBUG_ALL = log_world.info, log_world.debug
#
# Overlap with the the lat/lon box. The function only tests if there is a potential overlap.
# The default answer is that there is an overlap, unless we find it impossible.
#
def potoverlap(refpoly, testpoly) :
    po = True
    if np.max(np.array(testpoly)[:,0]) < np.min(np.array(refpoly)[:,0]) and np.min(np.array(testpoly)[:,0]) > np.max(np.array(refpoly)[:,0]) :
              po = False
    if np.max(np.array(testpoly)[:,1]) < np.min(np.array(refpoly)[:,1]) and np.min(np.array(testpoly)[:,1]) > np.max(np.array(refpoly)[:,1]) :
              po = False
    return po
#
# Routine to generate a square polygone around the centre point
#
@jit
def boxit(cent, dlon, dlat, polyres) :
    boxll=[]
    loninc = dlon/polyres
    latinc = dlat/polyres
    # Side 1
    for pts in range(polyres) :
        boxll.append([cent[0]-dlon/2.0+loninc*pts, cent[1]+dlat/2.0])
    # Side 2
    for pts in range(polyres) :
        boxll.append([cent[0]+dlon/2.0, cent[1]+dlat/2.0-latinc*pts])
    # Side 3
    for pts in range(polyres) :
        boxll.append([cent[0]+dlon/2.0-loninc*pts, cent[1]-dlat/2.0])
    # Side 4
    for pts in range(polyres) :
        boxll.append([cent[0]-dlon/2.0, cent[1]-dlat/2.0+latinc*pts])
    # Close
    boxll.append(boxll[0])
    return boxll
#
# Function to compute a distance in Lon/Lat space
#
@jit
def loladist(a,b) :
    return np.sqrt((a[0]-b[0])**2+(a[1]-b[1])**2)
#
# Function to compute maximum radius of cicle around polygon
#
@jit
def maxradius(cent, lon, lat) :
    radius=[]
    for lx, ly in zip(lon, lat) :
        radius.append(np.sqrt((cent[0]-lx)**2+(cent[1]-ly)**2))
    return max(radius)
#
# Simple routine to dump a field into a file.
#
def dumpfield(x, lon, lat, filename, varname) :
    #
    NCFillValue=1.0e20
    #
    print("Dumping overlap into file :", filename)
    i=np.nonzero(np.mean(x,axis=0))
    j=np.nonzero(np.mean(x,axis=1))
    ilen=np.max(i)-np.min(i)+1
    jlen=np.max(j)-np.min(j)+1
    print("Overlap")
    outnf=Dataset(filename, 'w', format='NETCDF4_CLASSIC')
    # Dimensions
    outnf.createDimension('lon', ilen)
    outnf.createDimension('lat', jlen)
    lonnc = outnf.createVariable("lon", vtyp, ('lat','lon'), fill_value=NCFillValue)
    lonnc.units="Longitude of grid box centre degrees east"
    lonnc.standard_name="grid longitude"
    lonnc.title="Longitude"
    lonnc.axis="X"
    lonnc[:,:] = lon[np.min(j):np.max(j)+1,np.min(i):np.max(i)+1]
    latnc = outnf.createVariable("lat", vtyp, ('lat','lon'), fill_value=NCFillValue)
    latnc.units="Latitude of grid box centre degrees north"
    latnc.standard_name="grid latitude"
    latnc.title="Latitude"
    latnc.axis="Y"
    latnc[:,:] = lat[np.min(j):np.max(j)+1,np.min(i):np.max(i)+1]
    #
    xnc = outnf.createVariable(varname, vtyp, ('lat','lon'), fill_value=NCFillValue)
    xnc[:,:] = x[np.min(j):np.max(j)+1,np.min(i):np.max(i)+1]
    outnf.close()
    #
    return
#
# Compute the weights of the overlap of modelgrif and hydrogrid
# Only if there is no file which already contains the information.
#
class compweights :
    def __init__(self, wfile, part, modelgrid, hydrogrid) :
        #
        sub_index = []
        sub_area = []
        sub_lon = []
        sub_lat = []
        nbpts = len(modelgrid.centers)
        #
        # Only compute the weights if the file does not exist.
        #
        if not os.path.exists(wfile) :
            INFO("Computing weights")
            nbpts = len(modelgrid.centers)
            i = 0
            for cell, center, radius, area in zip(modelgrid.polylist, modelgrid.centers, modelgrid.radius, modelgrid.area) :
                t = time.time()
                llinside=vector.vector_to_lonlat(cell.polygons[0].inside[0],cell.polygons[0].inside[1],cell.polygons[0].inside[2])
                area_in=np.zeros((hydrogrid.jjm,hydrogrid.iim), dtype=float)

                selected = hydrogrid.select(center, radius)
                for isel in selected :
                    hydrocell = hydrogrid.polylist[isel]
                    index = hydrogrid.index[isel]
                    if cell.intersects_poly(hydrocell):
                        inter = cell.intersection(hydrocell)
                        # Another strange behaviour of SphericalHarmonics. There should be an overlap but the intersection polygone is empty.
                        if len(inter.polygons) > 0 :
                            inside = inter.polygons[0]._inside
                            if hydrocell.contains_point(inside) and cell.contains_point(inside):
                                area_in[index[0],index[1]] = inter.area()*(EarthRadius**2)
                            else:
                                area_in[index[0],index[1]] = (4*np.pi-inter.area())*(EarthRadius**2)

                ##############
                # Output of Overlap areas for each grid point
                #
                #ncfile="Area_"+str("%.2f"%llinside[0])+"_"+str("%.2f"%llinside[1])+"_"+str(rank+1)+"of"+str(nbcore)+".nc"
                #RPP.dumpfield(area_in, hydrogrid.lon, hydrogrid.lat, ncfile, "Intersect area")
                #
                sub_index.append(np.array(np.where(area_in > 0.0)))
                sub_area.append(np.extract(area_in > 0.0, area_in))
                sub_lon.append(np.extract(area_in > 0.0, hydrogrid.lon))
                sub_lat.append(np.extract(area_in > 0.0, hydrogrid.lat))
                t1 = time.time()
                i += 1
                with open("./Weight_proc_{0}.txt".format(part.rank), "a") as foo:
                    foo.write("Pt : {0} / {1};  time: {2} s.\n".format(i, nbpts, int(t1-t)))
            with open("./Weight_proc_{0}.txt".format(part.rank), "a") as foo:
                foo.write("Finished !\n")
            #
            # Save information in class variables.
            #
            self.index = sub_index
            self.area = sub_area
            self.lon = sub_lon
            self.lat = sub_lat
            self.hpts = [l.shape[0] for l in self.lon]
            self.maxhpts = max(self.hpts)
            nboverlap=[]
            overlaperr=[]
            for icell in range(len(self.lon)) :
                nboverlap.append(self.hpts[icell])
                overlaperr.append(np.sum(self.area[icell])/modelgrid.area[icell])
            #
            INFO(str(part.rank)+" Area Overlap error : "+str(min(overlaperr))+"-"+str(max(overlaperr))+\
                     " Nb of Hydro grid overlaping : "+str(min(self.hpts))+"-"+str(max(self.hpts)) )
            #
            self.dumpweights(wfile, part, modelgrid, hydrogrid)
            #
            ##self.printtest([15,19], part, modelgrid)
            #
        else :
            #
            INFO("Reading weights from "+wfile)
            with open("check.out","a+") as foo:
               foo.write("Proc {0}, Start reading Weights from {1} points\n".format(part.rank,nbpts))
            t = time.time() 
            self.fetchweight(wfile, part, modelgrid, hydrogrid)
            self.hpts = [l.shape[0] for l in self.lon]
            self.maxhpts = max(self.hpts)
            t1 = time.time()
            INFO("Proc {0}, time : {1}".format(part.rank,round(t1-t,2 ))) 
            with open("check.out","a+") as foo:
               foo.write("Proc {0}, time : {1}\n".format(part.rank,round(t1-t,2 )))
            """
            for icell in range(len(self.lon)) :
                INFO(str(icell)+" Sum of overlap "+str(np.sum(self.area[icell])/modelgrid.area[icell])+
                     " Nb of Hydro grid overlaping : "+str(self.hpts[icell]))
            """
            #
            ##self.printtest([15,19], part, modelgrid)
            #
        return
    #
    # Function to fetch the weights from weight file
    #
    def fetchweight(self, weightfile, part, modelgrid, hydrogrid) :
        #
        debug_time = False
        #
        innf=Dataset(weightfile, 'r')
        #
        gindex=innf.variables["global_index"][:,:]
        indexfill = innf.variables["global_index"]._FillValue
        landind = part.toglobal_index(modelgrid.indP)
        #
        t = time.time()
        locinfile = [int(self.findinfile(pts, gindex)) for pts in landind]
        var = innf.variables["hydro_index"]
        self.hpts = [int(np.array(np.where(var[0,:,i]>= 0)).shape[1]) for i in locinfile ]
        A = np.where(np.array(self.hpts) == 0)[0]
        t1 = time.time()

        if debug_time:INFO("Proc {0}, step 1, time : {1}".format(part.rank,round(t1-t,2 )))
        #
        var = innf.variables["hydro_area"]
        self.area = [var[0:maxhpts,i] for i,maxhpts in zip(locinfile, self.hpts)]
        var = innf.variables["hydro_lon"]
        self.lon = [var[0:maxhpts,i] for i,maxhpts in zip(locinfile, self.hpts)]
        var = innf.variables["hydro_lat"]
        self.lat = [var[0:maxhpts,i] for i,maxhpts in zip(locinfile, self.hpts)]
        t2 = time.time()
        if debug_time: INFO("Proc {0}, step 2, time : {1}".format(part.rank,round(t2-t1,2 )))
        #
        innf.close()
        # Find the indicis of the points on the HydroGrid using the coordinates.
        self.index = self.findhindex(hydrogrid)
        for k in A:
            if modelgrid.centers[k][1]<-60:
                self.index[k] = np.array([[-1],[ -1]])
                self.area[k] = np.array([modelgrid.area[k]])
                self.lon[k] = np.array([modelgrid.centers[k][0]])
                self.lat[k] = np.array([modelgrid.centers[k][1]])
        t3 = time.time()
        if debug_time: INFO("Proc {0}, step 3, time : {1}".format(part.rank,round(t3-t2,2 )))
        #
        return
    #
    # Function to find the indicis of the points on the hydrological grid for the current
    # partition of the domain.
    #
    def findhindex(self,  hydrogrid): 
        ind=[self.subhindex(im, hydrogrid) for im in range(len(self.lon))]
        return ind

    def subhindex(self, im, hydrogrid ):
        lon_loc = [np.where(hydrogrid.lon[0,:] == self.lon[im][ih])[0][0] for ih in range(self.hpts[im])]
        lat_loc = [np.where(hydrogrid.lat[:,0] == self.lat[im][ih])[0][0] for ih in range(self.hpts[im])]
        pts_loc = [[a, b] for a,b in zip(lat_loc,lon_loc)]
        hind = np.transpose(pts_loc).astype(np.int32)
        return hind
    #
    # TEST function to be deleted later
    #
    def printtest(self, printpts, part, modelgrid) :
        landind = part.toglobal_index(modelgrid.indP)
        ip = IntFillValue
        for i in range(len(landind)) :
            if printpts[0] == landind[i][0] and printpts[1] == landind[i][1] :
                ip = i
        if ip >= 0 :
            print("H-Report on point ", landind[ip][:], " with ",self.index[ip].shape[1]," Hydro points.")
            print("H- indices : ", self.index[ip].astype(int))
            print("H- total area :", modelgrid.area[ip], np.sum(self.area[ip]), "Error =",modelgrid.area[ip]/np.sum(self.area[ip]))
            print("H- areas :", self.area[ip])
            print("H- lon :", self.lon[ip])
            print("H- lat :", self.lat[ip])
        return
    #
    # Function to find the index of the point in the file
    #
    def findinfile(self, pts, indlist) :
        ib = IntFillValue
        A = np.array([indlist[0,i] for i in range(indlist.shape[1])])
        B = np.array([indlist[1,i] for i in range(indlist.shape[1])])
        
        C = np.where((A==pts[0])*(B==pts[1]))[0]
        if len(C) ==1:
            ib = C[0]
        else:
            INFO("ERREUR C {0}".format(len(C)))
        return ib
    #
    # Function to dump all the weights into a netCDF file
    #
    def dumpweights(self, weightfile, part, modelgrid, hydrogrid) :
        #
        # This number needs to come from compweights
        #
        maxhydropts=part.domainmax(self.maxhpts)
        landind = part.toglobal_index(modelgrid.indP)
        #
        # On the root proc set-up the netCDF file with all the variables needed.
        #
        if part.rank == 0 :
            nf = self.generatenetcdffile(weightfile, part.nig, part.njg, part.nblandg, \
                                         maxhydropts, modelgrid.source, hydrogrid.source)
        #
        #
        #
        maxlandpoint = part.domainmax(len(part.landcorelist))
        print("Maximum points of Land points on any proc :", maxlandpoint, " Here : ", len(part.landcorelist), "Global Nbland = ", part.nblandg)
        #
        # Send each point on this proc to the roor_proc
        #
        ni = 0
        for i in range(maxlandpoint) :
            if i < len(part.landcorelist) :
                # If this is a core point then send it.
                glandind = part.gatherbypoint(landind[part.landcorelist[i]][:])
                ghindj = part.gatherbypoint(self.index[part.landcorelist[i]][0,:])
                ghindi = part.gatherbypoint(self.index[part.landcorelist[i]][1,:])
                ghlon = part.gatherbypoint(self.lon[part.landcorelist[i]][:])
                ghlat = part.gatherbypoint(self.lat[part.landcorelist[i]][:])
                gharea = part.gatherbypoint(self.area[part.landcorelist[i]][:])
            else :
                # If it is not a core point then send nothing to the root_proc
                glandind = part.gatherbypoint([])
                ghindi = part.gatherbypoint([])
                ghindj = part.gatherbypoint([])
                ghlon = part.gatherbypoint([])
                ghlat = part.gatherbypoint([])
                gharea = part.gatherbypoint([])
            #
            # On the root proc which collected all the data we write data to the netCDF file
            #
            if part.rank == 0 :
                ni = self.addtoncvar(ni, part.nig, part.njg, glandind, ghindi, ghindj, ghlon, ghlat, gharea)
                #
        if part.rank == 0 :
            nf.close()
        #
        return
    #
    # Function to generate the netCDF file for the weight functions
    #
    def generatenetcdffile(self, weightfile, nig, njg, nbland, maxhydropts, mfile, hfile) :
        #
        outnf=Dataset(weightfile, 'w', format='NETCDF4_CLASSIC')
        # Dimensions
        print("Generating dimensions : ", nig, njg, nbland, maxhydropts)
        outnf.createDimension('i', nig)
        outnf.createDimension('j', njg)
        outnf.createDimension('k', 2)
        outnf.createDimension('nbland', nbland)
        outnf.createDimension('nover', maxhydropts)
        # Global attributes
        outnf.HydroGrid=hfile
        outnf.ModelGrid=mfile
        #
        self.mask = outnf.createVariable("mask", vtyp, ('j','i'), fill_value=FillValue)
        self.mask.units="-"
        self.mask.standard_name="Land sea mask"
        self.mask.title="Land sea mask"
        self.mask[:,:] = 0
        #
        self.proc = outnf.createVariable("proc_num", vtyp, ('j','i'), fill_value=FillValue)
        self.proc.units="-"
        self.proc.standard_name="Processor working on land point"
        self.proc.title="Processor working on land point"
        self.proc[:,:] = FillValue
        #
        self.gindex = outnf.createVariable("global_index", vinttyp, ('k','nbland'), fill_value=IntFillValue)
        self.gindex.units="-"
        self.gindex.standard_name="Index j,i on Global Grid"
        self.gindex.title="Index j,i on Global Grid"
        self.gindex[:,:] = IntFillValue
        #
        self.hindex = outnf.createVariable("hydro_index", vinttyp, ('k','nover','nbland'), fill_value=IntFillValue)
        self.hindex.units="-"
        self.hindex.standard_name="Index of HydroGrid point contributing to current land point"
        self.hindex.title="Index of HydroGrid point contributing to current land point"
        self.hindex[:,:,:] = IntFillValue
        #
        self.hlon = outnf.createVariable("hydro_lon", vtyp, ('nover','nbland'), fill_value=FillValue)
        self.hlon.units="degrees_east"
        self.hlon.standard_name="Longitude of HydroGrid point contributing to current land point"
        self.hlon.title="Longitude of HydroGrid point contributing to current land point"
        self.hlon[:,:] = FillValue
        #
        self.hlat = outnf.createVariable("hydro_lat", vtyp, ('nover','nbland'), fill_value=FillValue)
        self.hlat.units="degrees_north"
        self.hlat.standard_name="Latitude of HydroGrid point contributing to current land point"
        self.hlat.title="Latitude of HydroGrid point contributing to current land point"
        self.hlat[:,:] = FillValue
        #
        self.harea = outnf.createVariable("hydro_area", vtyp, ('nover','nbland'), fill_value=FillValue)
        self.harea.units="m^2"
        self.harea.standard_name="Area of HydroGrid point contributing to current land point"
        self.harea.title="Area of HydroGrid point contributing to current land point"
        self.harea[:,:] = FillValue
        #
        return outnf
    #
    # Add variables
    #
    def addtoncvar(self, nib, nig, njg, landind, hindi, hindj, hlon, hlat, harea) :
        #
        mxval,nbp=landind.shape
        for ib in range(nbp) :
            if ~np.isnan(landind[0,ib]) :
                maxhpts = len(hlon[:,ib])
                j,i=landind[:,ib]
                self.mask[j,i] = 1.0
                self.proc[j,i] = ib
                self.gindex[0,nib] = j
                self.gindex[1,nib] = i
                self.hindex[0,:maxhpts,nib] = np.reshape(hindj[:,ib],self.hindex[0,:maxhpts,nib].shape)
                self.hindex[1,:maxhpts,nib] = np.reshape(hindi[:,ib],self.hindex[1,:maxhpts,nib].shape)
                self.hlon[:maxhpts,nib] = hlon[:,ib]
                self.hlat[:maxhpts,nib] = hlat[:,ib]
                self.harea[:maxhpts,nib] = harea[:,ib]

                nib = nib+1
        return nib
